--  2009.       GSI TECHNOLOGY
--                                              Jeff Daugherty
--                                              apps@gsitechnology.com
--  Version: 1.2
--
--      FileName: core_SIO_B2.vhd
--
-- Revision History
-- 06/04/08    1.1    1) Updated model to removed GSI library.
-- 07/21/09    1.2    1) Model now has a reduced memory core.  The size is controlled by A_BITS
-- 12/15/10    1.3    1) Model supports both reduced and actual density memory core.  The size is controlled by A_BITS
--
library ieee;
use ieee.std_logic_1164.all;
entity SIO_B2_VHDL_CORE is
  generic (
    constant A_size    : integer;
    constant mem_size  : integer;
    constant A_bits    : integer;
    constant DQ_width  : integer;
    constant MEM_STYLE : string(7 downto 1) := "NORMAL ");  -- REDUCED or NORMAL
  port (
    SIGNAL A         : IN  std_logic_vector(A_size-1 DOWNTO 0);  -- address
    SIGNAL D         : IN  std_logic_vector(DQ_width - 1 DOWNTO 0) ;  -- byte A data
    SIGNAL Q         : OUT std_logic_vector(DQ_width - 1 DOWNTO 0) ;  -- byte B data
    SIGNAL cBW       : IN  std_logic_vector(3 downto 0);   -- bank 1 write enable
    SIGNAL K         : IN  std_logic;   -- input clock
    SIGNAL nK        : IN  std_logic;   -- input clock bar
    SIGNAL C         : IN  std_logic;   -- output clock
    SIGNAL nC        : IN  std_logic;   -- output clock bar
    SIGNAL nLD       : IN  std_logic;   -- Read enable
    SIGNAL RnW       : IN  std_logic;   -- Write strobe
    SIGNAL nDoff     : IN  std_logic;   -- Write strobe
    SIGNAL TCK       : IN  std_logic;   -- Write strobe
    SIGNAL CQ        : OUT std_logic;   -- Write strobe
    SIGNAL nCQ       : OUT std_logic;   -- Write strobe
    SIGNAL HighZ     :     std_logic_vector(DQ_width-1 downto 0);
    SIGNAL tCHQV     :     time;
    SIGNAL tCHQX     :     time;
    SIGNAL tCQHQV    :     time;
    SIGNAL tCQHQX    :     time;
    SIGNAL tCHQV_off :     time;
    SIGNAL tCHQX_off :     time);
end SIO_B2_VHDL_CORE;

library ieee;
architecture SIO_B2 of SIO_B2_VHDL_CORE is
  use ieee.std_logic_unsigned.all, ieee.std_logic_arith.all;

  function pick (DQ_width: integer) return integer is
  begin
    if DQ_width = 18 or DQ_width = 8 then
      return 2;
    elsif DQ_width = 36 then
      return 4;
    else
      return 1;
    end if;
  end pick;

  type GSI_DATA_MEMORY is protected
    procedure set (variable address : integer;
                            data    : std_logic_vector(DQ_width - 1 downto 0);
                            bwi     : std_logic_vector(3 downto 0));
    procedure get (variable address : integer;
                   variable data    : out std_logic_vector(DQ_width - 1 downto 0));
  end protected GSI_DATA_MEMORY;

  type GSI_DATA_MEMORY is protected body
    type MEM is array (0 to 2**mem_size - 1) of std_logic_vector(DQ_width - 1 downto 0);
    variable MEMORY : MEM;
    procedure set (variable address : integer;
                            data    : std_logic_vector(DQ_width - 1 downto 0);
                            bwi     : std_logic_vector(3 downto 0)
                   ) is
      variable bw_size : integer := pick(DQ_width);
    begin
      for store in 0 to BW_size - 1 loop
        if bwi(store) = '1' then
          if DQ_width = 8 then
            MEMORY(address)(((store+1)*(DQ_width/2)) - 1 downto store*(DQ_width/2)) := data(((store+1)*(DQ_width/2)) - 1 downto store*(DQ_width/2));
          else
            MEMORY(address)(((store+1)*9) - 1 downto store*9) := data(((store+1)*9) - 1 downto store*9);
          end if;
        end if;
      end loop;
    end procedure set;
    procedure get(variable address : integer;
                  variable data : out std_logic_vector(DQ_width - 1 downto 0)
                  ) is
      begin
         data := MEMORY(address);
    end procedure get;
  end protected body GSI_DATA_MEMORY;
  
  type GSI_ADDR_MEMORY is protected
    procedure set (address : integer; variable data :     std_logic_vector(A_size downto 0));
    procedure get (address : integer; variable data : out std_logic_vector(A_size downto 0));
  end protected GSI_ADDR_MEMORY;

  type GSI_ADDR_MEMORY is protected body
    type MEM is array (0 to 2**A_bits) of std_logic_vector(A_size downto 0);
    variable MEMORY : MEM;
    procedure set (         address : integer;
                   variable data    : std_logic_vector(A_size downto 0)
                  ) is
      begin
        MEMORY(address) := data;
    end procedure set;
    procedure get(address : integer;
                  variable data : out std_logic_vector(A_size downto 0)
                  ) is
      begin
        data := MEMORY(address);
    end procedure get;
  end protected body GSI_ADDR_MEMORY;

  TYPE ADDRESS is array (0 to 7) of std_logic_vector(A_size DOWNTO 0);

  shared variable MEMORY_CORE    : GSI_DATA_MEMORY;
  shared variable ADDRESS_CORE   : GSI_ADDR_MEMORY;

  constant reset      : std_logic_vector(A_size downto 0) := conv_std_logic_vector(0, A_size+1);
  signal read_pos     : std_logic_vector(7 downto 0);
  signal Di           : std_logic_vector(DQ_width - 1 downto 0) ;
  signal Qi           : std_logic_vector(DQ_width - 1 downto 0) ;
  signal Qi_o         : std_logic_vector(DQ_width - 1 downto 0) ;
  signal bw_i         : std_logic_vector(3 downto 0);
  signal we           : std_logic := '0';
  signal we1          : std_logic := '0';
  signal re           : std_logic := '0';
  SIGNAL Q_switch     : std_logic := '0';
  SIGNAL DLL          : std_logic;
  SIGNAL K_i          : std_logic;
  SIGNAL CQ_o         : std_logic;
  SIGNAL nCQ_o        : std_logic;
  SIGNAL K_Time       : time      := 1 ns;
  SIGNAL nK_Time      : time      := 1 ns;
  SIGNAL CHQV         : time      := 1 ns;
  SIGNAL CHQX         : time      := 1 ns;
  SIGNAL CHCQon       : time      := 1 ns;
  SIGNAL CHCQoff      : time      := 1 ns;
  signal JTAG_Flag    : integer   := 0;
  signal K_Flag       : integer   := 0;
  signal C_Flag       : integer   := 0;
  signal Delay        : integer   := 0;
  signal Latency      : integer   := 0;
  signal Read_Command : string(4 downto 1);
  signal Write_Command: string(5 downto 1);

  signal wA_D       : ADDRESS := (reset, reset, reset, reset, reset, reset, reset, reset);
  signal rA_D       : ADDRESS := (reset, reset, reset, reset, reset, reset, reset, reset);
  
-- ---------------------------------------------------------------
--   	reduced memory stuff
-- ---------------------------------------------------------------
  signal address_used : std_logic_vector(a_bits-1 downto 0) := (others=>'0');
  constant MAX_BITS   : std_logic_vector(a_bits-1 downto 0) := (others=>'1');

--  TIMING FUNCTIONS
  function POSEDGE (SIGNAL s : std_ulogic) return BOOLEAN IS
  begin
    RETURN (s'EVENT AND ((To_X01(s'LAST_VALUE) = '0') OR (s = '1')));
  end; 

  function NEGEDGE (SIGNAL s : std_ulogic) return BOOLEAN IS
  begin
    RETURN (s'EVENT AND ((To_X01(s'LAST_VALUE) = '1') OR (s = '0')) );
  end; 
-- END TIMING FUNCTIONS

begin
  -- purpose: Main K controlled process
  -- inputs : K, nK, Di, bw_i, DLL, lA_B, 
  -- outputs: Qi, re, CQ_i, nCQ_i
  Main: process (K_i)

-- purpose: shift address arrays
-- inputs : wA_D, rA_D, read_pos, bw_i, write_data
-- outputs: wA_D, rA_D, read_pos, bw_i, write_data
    procedure shift is
      variable Temp_write : Address;
      variable Temp_read  : Address;
      variable Temp_read_pos : std_logic_vector(7 downto 0);
      variable i : integer;
    begin  -- shift
      Temp_write    := wA_D;
      Temp_read := rA_D;
      for i in 7 downto 1 loop
        Temp_write(i) := Temp_write(i-1);
        Temp_read(i) := Temp_read(i-1);
      end loop;  -- i

      wA_D       <= Temp_write;
      rA_D       <= Temp_read;
    end procedure shift;

-- purpose: Determine Read / Write / NOOP State
-- inputs : W, nW, R, nR, lA_B
-- outputs: Read_Command, Write_Command, Read_Pos, wA_D, rA_D, we
    procedure State is
    begin
      if(nLD='0' and RnW='1') then
        Read_Command <= "READ";
        re <= '1';
        rA_D(1) <= A & '0';
        rA_D(0) <= A & '1';
        we <= '0';
        Write_Command <= "NO-OP";
      elsif (nLD='0' and RnW='0') then
        we <= '1';
        Write_Command <= "Write";
        wA_D(1) <= A & '0';
        wA_D(0) <= A & '1';
        Read_Command <= "NOOP";
        re <= '0';
      else
        we <= '0';
        Write_Command <= "NO-OP";
        Read_Command <= "NOOP";
        re <= '0';
      end if; 
    end procedure State;

-- purpose: Read from array and set Qi
-- inputs : read_pos, re, r_count
-- outputs: re, Q_switch, Qi, r_count
    procedure Read_Array is
      variable address_fetch  : std_logic_vector(A_size downto 0);
      variable data_fetch     : std_logic_vector(DQ_width - 1 downto 0);
      variable index          : integer := 0;
      variable address_match  : integer := 0;
      variable address_index  : integer := 0;
    begin
      if MEM_STYLE = "REDUCED" then
        if conv_integer(address_used) > 0 then
          for index in 0 to conv_integer(address_used) - 1 loop
            ADDRESS_CORE.get(index, address_fetch);
            if (address_fetch = rA_D(latency) and address_match = 0) then
              address_match := 1;
              address_index := index;
            end if;
          end loop;  -- indev
        end if;
        if address_match = 1 then
          address_match := 0;
          MEMORY_CORE.get(address_index, data_fetch);
        else
          data_fetch := (others=>'X');
        end if;
      else
        address_index := conv_integer(rA_D(latency));
        MEMORY_CORE.get(address_index, data_fetch);
      end if;
      Qi <= data_fetch;
  end procedure Read_Array;

-- purpose: Write data to Array
-- inputs : we, bw_i, W
-- outputs: we1
    procedure Write_Array is
      variable addr_t         : integer;
      variable index          : integer := 0;
      variable address_match  : integer := 0;
      variable address_index  : integer := 0;
      variable address_fetch  : std_logic_vector(A_size downto 0);
      variable address_store  : std_logic_vector(A_size downto 0);
    begin
      we1 <= we;
      address_store := wA_D(2);
      if (we1='1')  then
        if MEM_STYLE = "REDUCED" then
          if conv_integer(address_used) >0 then
            for index in 0 to conv_integer(address_used)-1 loop
              ADDRESS_CORE.get(index, address_fetch);
              if (address_fetch = wA_D(2) and address_match=0) then
                address_match := 1;
                address_index := index;
              end if;
            end loop;  -- index
          end if;
          if address_match = 1 then
            address_match := 0;
            ADDRESS_CORE.set(address_index, address_store);
            addr_t := address_index;
          else
            if address_used = MAX_BITS then
              report "A Memory overflow has occurred.  Increase Size of A_BITS to increase memory size" severity failure;
            end if;
            ADDRESS_CORE.set(conv_integer(address_used), address_store);
            addr_t := conv_integer(address_used);
            address_used <= address_used + 1;
          end if;
          MEMORY_CORE.set(addr_t, Di, bw_i);
        else
          addr_t := conv_integer(wA_D(2));
          MEMORY_CORE.set(addr_t, Di, bw_i);
        end if;
      end if;
    end procedure Write_Array;

-- purpose: Main flow
    variable R : std_logic := '0';
  begin  -- process Main
    if DLL='1' then
      Delay   <=2;
      Latency <=3;
    else
      Delay   <=1;
      latency <=2;
    end if;

    if (posedge(K_i)) then
      shift;
      State;
      Write_Array;
      Read_Array;
    end if;

    if (negedge(K_i)) then
      shift;
      Write_Array;
      Read_Array;
    end if;

  end process Main;

-- purpose: Determine if TCK is running
-- inputs : TCK
-- outputs: JTAG_Flag
  JTAG : process (TCK)
    variable TCK_Count : integer := 0;
  begin  
    if (POSEDGE(TCK)) then
      if (TCK_Count>=2) then
        JTAG_Flag <= 1;
      end if;
      if (TCK_Count<2) then
        TCK_Count := TCK_Count + 1;
      end if;
    end if;
  end process JTAG;

  C_int : process (K, nK)
  begin  
    if (K = '1') then
      K_i  <= '1';
    elsif nK = '1' then
      K_i  <= '0';
    end if;
  end process C_int;

  C_nC : process (C, nC)
    variable C_Count   : integer := 0;
    variable nC_Count  : integer := 0;
  begin  
    if (POSEDGE(C)) then
      if (C_Count<2) then
        C_Count := C_Count + 1;
      end if;
    end if;
    if (POSEDGE(nC)) then
      if (nC_Count<2) then
        nC_Count := nC_Count + 1;
      end if;
    end if;
    if (C_Count>=2 and nC_Count>=2) then
      C_Flag <= 1;
    end if;
  end process C_nC;

  CQ_Clock : process (K, nK, C, nC)
    begin
      if K_flag = 1 then
        if ((C_Flag=0 and posedge(K)) or (C_Flag=1 and posedge(C))) then
          CQ_o  <= transport '1'after CHCQon;
          nCQ_o <= transport '0'after CHCQoff;
        end if;
        if ((C_Flag=0 and posedge(nK)) or (C_Flag=1 and posedge(nC))) then
          nCQ_o <= transport '1'after CHCQon;
          CQ_o  <= transport '0'after CHCQoff;
        end if;
      else
        CQ_o  <= 'Z';
        nCQ_o <= 'Z';
      end if;
    end process CQ_Clock;

  Q_Clock : process (CQ_o, nCQ_o)
    begin
      if posedge(CQ_o) or posedge(nCQ_o) then
        for i in 7 downto 1 loop
          read_pos(i) <= read_pos(i-1);
        end loop;  -- i
        read_pos(0) <= re;
        if (read_pos(Delay) = '1') then
          Q_switch <= transport '1' after (CHQV);
          Q_switch <= transport '0' after (K_time + CHQX);
        else
          Q_switch <= transport '0' after (K_time + CHQX);
        end if;
      end if;
      Qi_o <= Qi;
    end process Q_Clock;

-- purpose: Find time since last K clock edge
-- inputs : K
-- outputs: K_time
  time_K_clk : process (K)
    variable K_Count   : integer := 0;
  begin
    if K_count < 4  then
      K_count := K_count + 1;
      K_time <= 1.0 ns;
    else
      K_time <= K'delayed'last_event;
    end if;
  end process time_K_clk;

-- purpose: Find time since last K clock edge
-- inputs : K
-- outputs: K_time
  time_nK_clk : process (nK)
    variable nK_Count   : integer := 0;
  begin
    if nK_count < 4  then
      nK_count := nK_count + 1;
      nK_time <= 1.0 ns;
    else
      K_flag <= 1;
      nK_time <= nK'delayed'last_event;
    end if;
  end process time_nK_clk;

  Di <= D;
  bw_i <= not cBW;
  CHQV <= tCHQV when DLL = '1' else tCHQV_off;
  CHQX <= tCHQX when DLL = '1' else tCHQX_off;
  CHCQon  <= 4*K_time + (CHQV - tCQHQV);
  CHCQoff <= 4*K_time + (CHQV - tCQHQV);
--  CHCQoff <= 4*K_time + (CHQX - tCQHQX);
  DLL  <= nDoff;
  CQ   <= CQ_o;
  nCQ  <= nCQ_o;
  Q    <= Qi_o when Q_switch = '1' else (others=>'Z');
end SIO_B2;
